/*
oStorybook: Open Source software for novelists and authors.
initial conception Martin Mustun, adaptation FaVdB

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a fileCopy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package storybook;

import assistant.Assistant;
import assistant.app.AssistantApp;
import i18n.I18N;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.HeadlessException;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.beans.PropertyChangeEvent;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileLock;
import java.nio.file.Files;
import java.nio.file.Path;
import static java.nio.file.StandardCopyOption.REPLACE_EXISTING;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import org.hibernate.Session;
import resources.icons.ICONS;
import resources.icons.IconUtil;
import storybook.Const.STORYBOOK;
import storybook.exim.importer.ImportDocument;
import storybook.model.H2File;
import storybook.model.book.Book;
import storybook.model.book.BookUtil;
import storybook.model.oldmodel.ModelMigration;
import storybook.toolkit.LOG;
import storybook.toolkit.StringUtil;
import storybook.toolkit.exec.CallProcess;
import storybook.toolkit.file.EnvUtil;
import static storybook.toolkit.file.EnvUtil.getHomeDir;
import storybook.toolkit.file.IOUtil;
import storybook.toolkit.net.Updater;
import storybook.toolkit.swing.FontUtil;
import storybook.toolkit.swing.LaF;
import storybook.toolkit.swing.SwingUtil;
import storybook.toolkit.swing.splash.WaitingSplash;
import storybook.ui.MainFrame;
import storybook.ui.dialog.ConfirmDlg;
import storybook.ui.dialog.ExceptionDlg;
import storybook.ui.dialog.FirstStartDlg;
import storybook.ui.dialog.MessageDlg;
import storybook.ui.dialog.NewProjectDlg;
import storybook.ui.dialog.preferences.PropertiesDlg;

public class App extends Component implements ClipboardOwner {

	private static final String TT = "App";

	public static Color getDefaultForeground() {
		JTextField tf = new JTextField();
		return tf.getForeground();
	}

	public Pref preferences;
	private final List<MainFrame> mainFrames;
	private Font fontDefault,
			fontEditor,
			fontMono,
			fontBook;

	private static App instance;
	private static String i18nFile = "";
	private static boolean dev = false;

	public static boolean isDev() {
		return dev;
	}

	private Assistant assistant;
	private static boolean python;
	private static boolean dictaphone = false, dictaphoneStarted = false;

	private App() {
		mainFrames = new ArrayList<>();
		//init();
	}

	public static Pref getPref() {
		return (getInstance().preferences);
	}

	private void init() {
		//LOG.trace(TT + ".init()");
		try {
			initClipboard();
		} catch (Exception ex) {
			//empty
		}
		preferences = new Pref();
		if (preferences.getBoolean(Pref.KEY.FIRST_START, true)) {
			preferences.initFont();
		}
		LaF.set();
		assistant = new Assistant();
		String s = preferences.getString(Pref.KEY.ASSISTANT);
		if (!s.isEmpty()) {
			setAssistant(s);
		}
		initI18N();
		IconUtil.setDefSize();
		fontRestoreDefault();
	}

	private void init(File fileToOpen) {
		//LOG.trace(TT + ".init(fileToOpen=" + (fileToOpen != null ? fileToOpen.getAbsolutePath() : "null") + ")");
		init();
		dictaphoneInit();
		try {
			// first start dialog
			if (preferences.getBoolean(Pref.KEY.FIRST_START, true)) {
				LOG.trace("call First Start Dialog");
				FirstStartDlg dlg = new FirstStartDlg();
				dlg.setVisible(true);
				preferences.setBoolean(Pref.KEY.FIRST_START, false);
			}

			boolean fileHasBeenOpened = false;
			H2File dbFile;
			if (fileToOpen != null) {
				LOG.trace("trying to open " + fileToOpen.getAbsolutePath());
				dbFile = new H2File(fileToOpen.getAbsolutePath());
				fileHasBeenOpened = openFile(dbFile, true);
			} else if (preferences.getBoolean(Pref.KEY.OPEN_LASTFILE, false)
					&& !preferences.getString(Pref.KEY.LASTOPEN_FILE, "").isEmpty()) {
				String str = preferences.getString(Pref.KEY.LASTOPEN_FILE, "");
				if (str.endsWith("h2.db")) {
					str = str.replace("h2.db", "osbk");
				}
				if (str.endsWith("mv.db")) {
					str = str.replace("mv.db", "osbk");
				}
				dbFile = new H2File(str);
				fileHasBeenOpened = openFile(dbFile, true);
			}
			if (!fileHasBeenOpened) {
				preferences.setString(Pref.KEY.LASTOPEN_FILE, "");
			}
			MainFrame mainFrame = new MainFrame();
			mainFrame.init();
			mainFrame.initBlankUi();
			addMainFrame(mainFrame);
			// check for updates
			Updater.checkForUpdate(false);
		} catch (Exception e) {
			LOG.err(TT + ".init()", e);
			ExceptionDlg.show("", e);
		}
	}

	public void initI18N() {
		//LOG.trace("App.initI18N()");
		Locale locale = Locale.getDefault();
		try {
			String localeStr = preferences.getString(Pref.KEY.LANGUAGE, "en_US");
			String language[] = localeStr.split("_");
			locale = new Locale(language[0], language[1]);
			setLocale(locale);
		} catch (Exception ex) {
			LOG.err("unable to find Locale in oStorybook.ini");
		}
		try {
			String f = preferences.getString(Pref.KEY.MSGFILE, "");
			if (!f.isEmpty()) {
				I18N.setFileMessages(f);
			}
		} catch (Exception ex) {
		}
		I18N.initMessages(locale);
	}

	public static Assistant getAssistant() {
		return getInstance().assistant;
	}

	public void setAssistant(String nf) {
		assistant.init(nf);
	}

	public List<MainFrame> getMainFrames() {
		return mainFrames;
	}

	public void addMainFrame(MainFrame mainFrame) {
		//LOG.trace("App.addMainFrame("+mainFrame.getName()+")");
		mainFrames.add(mainFrame);
	}

	public void removeMainFrame(MainFrame mainFrame) {
		//LOG.trace("App.removeMainFrame("+mainFrame.getName()+")");
		if (mainFrame.getH2File() != null) {
			mainFrame.getH2File().remove();
		}
		mainFrames.remove(mainFrame);
	}

	public void closeBlank() {
		//LOG.trace("App.closeBlank()");
		for (MainFrame mainFrame : mainFrames) {
			if (mainFrame.isBlank()) {
				mainFrames.remove(mainFrame);
				mainFrame.dispose();
			}
		}
	}

	public static App getInstance() {
		if (instance == null) {
			instance = new App();
		}
		return instance;
	}

	public static String getFileName(String thetitle) {
		//LOG.trace("App.createNewFile()");
		String titlename = StringUtil.escapeTxt(thetitle);
		String x[] = titlename.split(" ");
		titlename = "";
		for (String z : x) {
			titlename += StringUtil.capitalize(z);
		}
		String filename = EnvUtil.getHomeDir().getPath()
				+ File.separator
				+ titlename
				+ STORYBOOK.FILE_EXT_OSBK.toString();
		File f = IOUtil.fileSelect(null, filename, "osbk", "osbk", "file.create");
		if (f == null) {
			return "";
		}
		return f.getAbsolutePath();
	}

	private MainFrame initNewFile(H2File dbFile, String thetitle, String... fromDoc) {
		//LOG.trace(TT + ".initNewFile(dbFile=" + dbFile.getH2Name() + ", title=" + thetitle + ")");
		MainFrame newMainFrame = new MainFrame();
		newMainFrame.init(dbFile);
		//initNewProperties(newMainFrame, thetitle, fromDoc);
		return newMainFrame;
	}

	private void initNewProperties(MainFrame newMainFrame, String thetitle, String... fromDoc) {
		PropertiesDlg properties = new PropertiesDlg(newMainFrame, true);
		properties.tfTitle.setText(thetitle);
		if (fromDoc != null && fromDoc.length > 0) {
			String ext = (fromDoc[0].endsWith(".docx") ? "docx" : "odt");
			properties.getXeditor().setExtension(ext);
		}
		properties.setVisible(true);
		newMainFrame.getBook().setProperties(properties);
		newMainFrame.getBook().setString(Book.PARAM.LAYOUT_BOOK,
				preferences.getString(Pref.KEY.BOOK_LAYOUT));
	}

	private void initNewFileEnd(MainFrame newMainFrame, H2File dbFile, String thetitle) {
		//LOG.trace(TT + ".initNewFileEnd(mainFrame, dbFile=" + dbFile.getH2Name() + ", title=" + thetitle + ")");
		newMainFrame.initUi();
		newMainFrame.getBookController().fireAgain();
		addMainFrame(newMainFrame);
		closeBlank();
		updateFilePref(dbFile, thetitle);
		newMainFrame.reloadBook();
		newMainFrame.saveFile(true);
		//create new Images directory
		IOUtil.directoryCreate(dbFile.getPath(), "Images");
		//create new Documents directory
		IOUtil.directoryCreate(dbFile.getPath(), "Documents");
		initNewProperties(newMainFrame, thetitle);
		setDefaultCursor();
	}

	public void createNewFile() {
		LOG.trace(TT + ".createNewFile()");
		NewProjectDlg dlg = new NewProjectDlg(mainFrames.get(0));
		dlg.setVisible(true);
		if (dlg.isCanceled()) {
			return;
		}
		String thetitle = dlg.getTitle();
		int nature = dlg.getNature();
		int nbParts = dlg.getNbParts();
		int nbChapters = dlg.getNbChapters();
		Date objective = dlg.getObjective();
		String filename = getFileName(thetitle);
		if (filename.isEmpty()) {
			return;
		}
		if (filename.endsWith(STORYBOOK.FILE_EXT_H2DB.toString())) {
			filename = filename.replace(STORYBOOK.FILE_EXT_H2DB.toString(), "");
		}
		if (filename.endsWith(STORYBOOK.FILE_EXT_MVDB.toString())) {
			filename = filename.replace(STORYBOOK.FILE_EXT_MVDB.toString(), "");
		}
		if (filename.endsWith(STORYBOOK.FILE_EXT_OSBK.toString())) {
			filename = filename.replace(STORYBOOK.FILE_EXT_OSBK.toString(), "");
		}
		File fOsbk = new File(filename + STORYBOOK.FILE_EXT_OSBK.toString());
		File fH2db = new File(filename + STORYBOOK.FILE_EXT_H2DB.toString());
		File fMvdb = new File(filename + STORYBOOK.FILE_EXT_MVDB.toString());
		if (fOsbk.exists() || fH2db.exists() || fMvdb.exists()) {
			int ret = ConfirmDlg.show(null,
					I18N.getMsg("file.save.overwrite.title"),
					I18N.getMsg("file.exists", filename),
					false);
			if (ret == ConfirmDlg.CANCEL) {
				return;
			}
			IOUtil.fileDelete(fOsbk);
			IOUtil.fileDelete(fH2db);
			IOUtil.fileDelete(fMvdb);
		}
		LOG.trace("create new file");
		H2File dbFile = new H2File(filename + STORYBOOK.FILE_EXT_MVDB.toString());
		String dbName = dbFile.getH2Name();
		if (dbName == null) {
			return;
		}
		MainFrame newMainFrame = initNewFile(dbFile, thetitle);
		newMainFrame.getBookModel().initEntities(nature, nbParts, nbChapters, objective);
		initNewFileEnd(newMainFrame, dbFile, thetitle);
	}

	/**
	 * import the document in OSBK
	 *
	 * @param hi
	 */
	public void importBook(ImportDocument hi) {
		//LOG.trace(TT + ".importBook(hi)");
		App app = App.getInstance();
		MainFrame newMainFrame = initNewFile(hi.getDBFile(), hi.getTitle(), hi.getFile().getAbsolutePath());

		newMainFrame.getBookModel().initBaseEntities(hi.getTitle());
		newMainFrame.getBook().setString(Book.PARAM.LAYOUT_SCENE_SEPARATOR,
				app.preferences.getString(Pref.KEY.SCENE_SEPARATOR));

		Session session = newMainFrame.getBookModel().beginTransaction();
		hi.getAll(session);
		session.getTransaction().commit();

		initNewFileEnd(newMainFrame, hi.getDBFile(), hi.getTitle());
	}

	/**
	 * rename the OSBK file
	 *
	 * @param mainFrame
	 * @param outFile
	 */
	public void renameFile(final MainFrame mainFrame, File outFile) {
		//LOG.trace("App.renameFile(mainFrame,outFile="+outFile.getAbsolutePath()+")");
		try {
			File inFile = mainFrame.getH2File().getFile();
			mainFrame.close(false);
			Path inPath = inFile.toPath();
			Path outPath = outFile.toPath();
			Files.move(inPath, outPath, REPLACE_EXISTING);
			H2File dbFile = new H2File(outFile);
			openFile(dbFile, true);
		} catch (IOException e) {
			LOG.err(TT + ".renameFile(" + mainFrame.getName() + "," + outFile.getName() + ")", e);
		}
	}

	/**
	 * open the project file
	 *
	 * @return
	 */
	public boolean openFile() {
		//LOG.trace("App.openFile()");
		final H2File h2File = BookUtil.openDocumentDialog();
		if (h2File == null) {
			return false;
		}
		return openFile(h2File, true);
	}

	/**
	 * open the DB file
	 *
	 * @param h2File
	 * @param oldPath
	 * @param newPath
	 * @return
	 */
	public boolean openFile(final H2File h2File, String oldPath, String newPath) {
		boolean rc = openFile(h2File, true);
		SwingUtilities.invokeLater(() -> App.changingDbFile(h2File, oldPath, newPath));
		return rc;
	}

	/**
	 * change the file
	 *
	 * @param h2File
	 * @param oldPath
	 * @param newPath
	 */
	public static void changingDbFile(H2File h2File, String oldPath, String newPath) {
		MainFrame mf = null;
		for (MainFrame m : App.getInstance().mainFrames) {
			LOG.trace("changingDbFile h2File=" + h2File.getH2Name() + ", m.dbFile=" + m.getH2File().getH2Name());
			if (m.getH2File().equals(h2File)) {
				mf = m;
			}
		}
		if (mf == null) {
			LOG.trace("changingDbFile mf=null");
			return;
		}
		mf.changePath(oldPath, newPath);
		mf.getBook().setTitle(I18N.getMsg("copyof") + " " + mf.getBook().getTitle());
	}

	/**
	 * open the DB file
	 *
	 * @param h2File
	 * @param mustExists
	 * @return
	 */
	public boolean openFile(H2File h2File, boolean mustExists) {
		//LOG.trace("App.openFile(" + h2File.getH2Name() + ")");
		if (!h2File.isOK() || !h2File.open()) {
			return (false);
		}
		if (h2File.isAlreadyOpened()) {
			return (true);
		}
		try {
			// model update from STORYBOOK 3.x to 4.0
			final ModelMigration oldPersMngr = ModelMigration.getInstance();
			if (oldPersMngr.open(h2File)) {
				try {
					if (!oldPersMngr.checkAndAlterModel(false)) {
						oldPersMngr.closeConnection();
						return false;
					}
				} catch (Exception e) {
					oldPersMngr.closeConnection();
					LOG.err("App.openFile(" + h2File.getH2Name() + ")", e);
					ExceptionDlg.show("", e);
					return false;
				}
				oldPersMngr.closeConnection();
				if (oldPersMngr.isAltered()) {
					MessageDlg.show(null,
							I18N.getMsg("file.optimize_warning"),
							I18N.getMsg("file.optimize"),
							true);
				}
			}
			setWaitCursor();
			final WaitingSplash dlg = new WaitingSplash(null, I18N.getMsg("loading", h2File.getName()));
			SwingUtilities.invokeLater(() -> {
				try {
					setWaitCursor();
					//trace("loading file... " + h2File);
					dlg.setText("Initialize MainFrame...");
					MainFrame newMainFrame = new MainFrame(h2File);
					dlg.setText("Adding MainFrame...");
					addMainFrame(newMainFrame);
					dlg.setText("Close blank...");
					closeBlank();
					dlg.setText("Update preferences...");
					updateFilePref(h2File, newMainFrame.getBook().getTitle());
					dlg.setText("Reload menu bars...");
					reloadMenuBars();
					setDefaultCursor();
				} catch (Exception e) {
					ExceptionDlg.show("", e);
				}
				dlg.dispose();
			});
		} catch (HeadlessException e) {
			LOG.err("App.openFile", e);
		}
		return true;
	}

	/**
	 * update the file preferences
	 *
	 * @param h2File
	 * @param title
	 */
	public void updateFilePref(H2File h2File, String title) {
		//LOG.trace("App.updateFilePref(" + h2File.getH2Name() + ")");
		if (h2File == null || title == null || title.isEmpty()) {
			return;
		}
		// save last openDocument directory and file
		File file = h2File.getFile();
		preferences.setString(Pref.KEY.LASTOPEN_DIR, file.getParent());
		preferences.setString(Pref.KEY.LASTOPEN_FILE, file.getPath());
		// save recent files
		preferences.recentFilesAdd(h2File.getOSBK(), title);
		preferences.save();
		reloadMenuBars();
	}

	/**
	 * clear the recent files list
	 *
	 */
	public void clearRecentFiles() {
		//LOG.trace("App.clearRecentFiles()");
		preferences.recentFilesClear();
		reloadMenuBars();
	}

	/**
	 * exit the application
	 *
	 */
	public void exit() {
		//LOG.trace("App.exit()");
		if (!mainFrames.isEmpty()
				&& preferences.getBoolean(Pref.KEY.CONFIRM_EXIT)
				&& ConfirmDlg.show(mainFrames.get(0),
						I18N.getMsg("exit"),
						I18N.getMsg("want.exit"),
						false) == ConfirmDlg.CANCEL) {
			return;

		}
		preferences.save();
		System.exit(0);
	}

	/**
	 * reset the font to default
	 *
	 */
	public void fontResetUi() {
		if (fontDefault == null) {
			return;
		}
		FontUtil.setDefaultFont(fontDefault);
	}

	/**
	 * set the default Font
	 *
	 * @param font
	 */
	public void fontSetDefault(Font font) {
		//LOG.trace("App.fontSetDefault(font="+(font!=null?font.toString():"null")+")");
		if (font == null) {
			JMenu e = new JMenu();
			fontDefault = e.getFont();
		}
		fontDefault = font;
		preferences.setString(Pref.KEY.FONT_DEFAULT, FontUtil.getString(fontDefault));
		fontResetUi();
	}

	/**
	 * get the default font
	 *
	 * @return
	 */
	public Font fontGetDefault() {
		if (fontDefault == null) {
			fontRestoreDefault();
		}
		return this.fontDefault;
	}

	/**
	 * restore the default font
	 *
	 */
	public void fontRestoreDefault() {
		//LOG.trace("App.fontRestoreDefault()");
		if (preferences == null) {
			preferences = new Pref();
		}
		String name = preferences.getString(Pref.KEY.FONT_DEFAULT, Const.FONT_DEFAULT);
		fontSetDefault(FontUtil.getFont(name));
	}

	/**
	 * set the Book font
	 *
	 * @param font
	 */
	public void fontSetBook(Font font) {
		//LOG.trace("App.fontSetBook()");
		if (font == null) {
			JEditorPane e = new JEditorPane();
			fontBook = e.getFont();
		} else {
			fontBook = font;
		}
		preferences.setString(Pref.KEY.FONT_BOOK, FontUtil.getString(fontBook));
	}

	/**
	 * get the Book font
	 *
	 * @return
	 */
	public Font fontGetBook() {
		if (fontBook == null) {
			fontRestoreBook();
		}
		return this.fontBook;
	}

	/**
	 * restore the Book font
	 *
	 */
	public void fontRestoreBook() {
		//LOG.trace("App.fontRestoreBook()");
		String name = preferences.getString(Pref.KEY.FONT_BOOK, Const.FONT_BOOK);
		fontSetBook(FontUtil.getFont(name));
	}

	/**
	 * set the editor font
	 *
	 * @param font
	 */
	public void fontSetEditor(Font font) {
		//trace("App.fontSetEditor()");
		if (font == null) {
			JEditorPane e = new JEditorPane();
			fontEditor = e.getFont();
		} else {
			fontEditor = font;
		}
		preferences.setString(Pref.KEY.FONT_EDITOR, FontUtil.getString(fontEditor));
	}

	/**
	 * get the editor font
	 *
	 * @return
	 */
	public Font fontGetEditor() {
		if (fontEditor == null) {
			fontRestoreEditor();
		}
		return this.fontEditor;
	}

	/**
	 * restore the editor font
	 *
	 */
	public void fontRestoreEditor() {
		//trace("App.fontRestoreEditor()");
		String name = preferences.getString(Pref.KEY.FONT_EDITOR, Const.FONT_EDITOR);
		fontSetEditor(FontUtil.getFont(name));
	}

	/**
	 * set the Monospaced font
	 *
	 */
	public void fontSetMono(Font font) {
		//LOG.trace("App.fontSetMono()");
		if (font == null) {
			JEditorPane e = new JEditorPane();
			fontMono = e.getFont();
		} else {
			fontMono = font;
		}
		preferences.setString(Pref.KEY.FONT_MONO, FontUtil.getString(fontMono));
	}

	/**
	 * get the Monospaced font
	 *
	 * @return
	 */
	public Font fontGetMono() {
		if (fontMono == null) {
			fontRestoreMono();
		}
		return this.fontMono;
	}

	/**
	 * restore the Monospaced font
	 *
	 */
	public void fontRestoreMono() {
		//LOG.trace("App.fontRestoreMono()");
		String name = preferences.getString(Pref.KEY.FONT_MONO, Const.FONT_MONO);
		fontSetMono(FontUtil.getFont(name));
	}

	/**
	 * refresh the data
	 *
	 */
	public void refresh() {
		//LOG.trace(TT+".refresh()");
		for (MainFrame mainFrame : mainFrames) {
			int width = mainFrame.getWidth();
			int height = mainFrame.getHeight();
			boolean maximized = mainFrame.isMaximized();
			mainFrame.getSbActionManager().reloadMenuToolbar();
			mainFrame.setSize(width, height);
			if (maximized) {
				mainFrame.setMaximized();
			}
			mainFrame.refresh();
		}
	}

	/**
	 * reload the menu bar
	 *
	 */
	public void reloadMenuBars() {
		for (MainFrame mainFrame : mainFrames) {
			mainFrame.getSbActionManager().reloadRecentMenu();
			mainFrame.getSbActionManager().reloadMenuToolbar();
			mainFrame.getSbActionManager().reloadWindowMenu();
		}
	}

	/**
	 * reload the status bar
	 *
	 */
	public void reloadStatusBars() {
		for (MainFrame mainFrame : mainFrames) {
			mainFrame.refreshStatusBar();
		}
	}

	/**
	 * set cursor waiting
	 *
	 */
	public void setWaitCursor() {
		for (MainFrame mainFrame : mainFrames) {
			SwingUtil.setWaitingCursor(mainFrame);
		}
	}

	/**
	 * set default cursor
	 *
	 */
	public void setDefaultCursor() {
		for (MainFrame mainFrame : mainFrames) {
			SwingUtil.setDefaultCursor(mainFrame);
		}
	}

	/**
	 * action for model properties change
	 *
	 * @param evt
	 */
	public void modelPropertyChange(PropertyChangeEvent evt) {
		// works, but currently not used
		// may be used for entity copying between files
	}

	static File fileToOpen = null;

	/**
	 * main process
	 *
	 * @param args
	 */
	public static void main(String[] args) {
		LOG.init(false, false);
		String vargs = String.join(" ", args);
		if (vargs.length() == 0) {
			vargs = "no option";
		}
		LOG.log(Const.getFullName() + " starting with: " + vargs);
		String tempDir = System.getProperty("java.io.tmpdir");
		String fn = tempDir + File.separator + "storybook.lck";
		if (args.length > 0) {
			for (int i = 0; i < args.length; i++) {
				switch (args[i].toLowerCase()) {
					case "--assistant":
						// start the AssistantApp
						SwingUtilities.invokeLater(() -> {
							AssistantApp.main(args);
						});
						return;
					case "--trace":
						LOG.log("Set trace mode");
						LOG.setTrace();
						break;
					case "--hibernate":
						LOG.log("Trace Hibernate");
						LOG.setTraceHibernate();
						break;
					case "--dev":
						App.dev = true;
						LOG.log("Development test");
						break;
					case "--msg":
						i++;
						String fileI18N = args[i];
						File f = new File(args[i]);
						if (!f.exists()) {
							fileI18N = args[i] + ".properties";
							f = new File(fileI18N);
							if (!f.exists()) {
								LOG.log("Msg test file not exists : " + fileI18N);
								fileI18N = "";
							}
						}
						if (!fileI18N.isEmpty()) {
							App.i18nFile = fileI18N;
							LOG.log("Message will be : " + fileI18N);
							I18N.setFileMessages(i18nFile);
						}
						break;
					default:
						//perhaps the file to open
						fileToOpen = new File(args[i]);
						if (fileToOpen.exists() && fileToOpen.getAbsolutePath().endsWith(".osbk")) {
							/*H2File dbFile = new H2File(fileToOpen.getAbsoluteFile());
							fileHasBeenOpened = openFile(dbFile, true);*/
						} else {
							fileToOpen = null;
						}
						break;
				}
			}
		}
		if (!lockInstance(fn)) {
			Object[] options = {I18N.getMsg("running.remove"), I18N.getMsg("cancel")};
			int n = JOptionPane.showOptionDialog(null,
					I18N.getMsg("running.msg"),
					I18N.getMsg("running.title"),
					JOptionPane.YES_NO_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE, null, options, options[1]);
			if (n == 1) {
				File file = new File(fn);
				if (file.exists() && file.canWrite() && !file.delete()) {
					JOptionPane.showMessageDialog(null, "Delete failed",
							"File\n" + file.getAbsolutePath() + "\ncould not be deleted.",
							JOptionPane.ERROR_MESSAGE);
				}
			}
			return;
		}

		python = CallProcess.initPython();
		SwingUtilities.invokeLater(() -> {
			App app = App.getInstance();
			app.init(fileToOpen);
		});
	}

	/**
	 * lock the instance
	 *
	 * @param lockFile
	 * @return
	 */
	private static boolean lockInstance(final String lockFile) {
		try {
			final File file = new File(lockFile);
			final RandomAccessFile randomAccessFile = new RandomAccessFile(file, "rw");
			final FileLock fileLock = randomAccessFile.getChannel().tryLock();
			if (fileLock != null) {
				Runtime.getRuntime().addShutdownHook(new Thread() {
					@Override
					public void run() {
						try {
							fileLock.release();
							randomAccessFile.close();
							file.delete();
						} catch (IOException e) {
							LOG.err("Unable to remove lock file: " + lockFile, e);
						}
					}
				});
				return true;
			}
		} catch (IOException e) {
			LOG.err("Unable to create and/or lock file: " + lockFile, e);
		}
		return false;
	}

	/**
	 * get the I18N file
	 *
	 * @return
	 */
	public String getI18nFile() {
		return (i18nFile);
	}

	/**
	 * set the I18N file
	 *
	 * @param file
	 */
	public void setI18nFile(String file) {
		i18nFile = file;
	}

	/**
	 * refresh all main frames
	 *
	 */
	public void refreshViews() {
		//LOG.trace("App.refreshViews()");
		for (MainFrame mainFrame : mainFrames) {
			mainFrame.refresh();
		}
	}

	/**
	 * clipboard changes
	 *
	 * @param clipboard
	 * @param contents
	 */
	@Override
	public void lostOwnership(Clipboard clipboard, Transferable contents) {
		//LOG.trace("App.lostOwnership");
		for (MainFrame mainFrame : mainFrames) {
			mainFrame.getSbActionManager().enablePaste();
		}
	}

	public boolean hasPasted = false;

	/**
	 * init clipboard changes
	 *
	 */
	private void initClipboard() {
		try {
			Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
			Transferable contents = clipboard.getContents(null);
			hasPasted = (contents != null) && contents.isDataFlavorSupported(DataFlavor.stringFlavor);
			if (hasPasted) {
				String result = (String) contents.getTransferData(DataFlavor.stringFlavor);
				StringSelection stringSelection = new StringSelection(result);
				clipboard.setContents(stringSelection, this);
			}
		} catch (HeadlessException | UnsupportedFlavorException | IOException ex) {
			//ex.printStackTrace(System.err);
		}
	}

	/**
	 * refresh the status bar
	 *
	 */
	public void refreshStatusBar() {
		for (MainFrame m : mainFrames) {
			m.refreshStatusBar();
		}
	}

	/**
	 * initialize the dictaphone
	 *
	 */
	private static void dictaphoneInit() {
		//LOG.trace(TT + ".initDictaphone()");
		File f = new File(getHomeDir().getAbsolutePath()
				+ File.separator + ".storybook5/dictaphone");
		dictaphone = f.exists();
	}

	/**
	 * check if dictaphone is installed
	 *
	 * @return
	 */
	public boolean dictaphoneHas() {
		if (!isDev()) {
			return false;
		}
		return dictaphone;
	}

	/**
	 * check if the dictaphone is started
	 *
	 * @return
	 */
	public boolean dictaphoneIsStarted() {
		return dictaphoneStarted;
	}

	/**
	 * start/stop the dictaphone
	 *
	 * @param it
	 */
	public static void dictaphoneStart(JMenuItem it) {
		if (dictaphoneStarted) {
			dictaphoneStarted = false;
			it.setText(I18N.getMsg("dictaphone.start"));
			it.setIcon(IconUtil.getIconSmall(ICONS.K.DICTAPHONE_START));
		} else {
			dictaphoneStarted = true;
			it.setText(I18N.getMsg("dictaphone.stop"));
			it.setIcon(IconUtil.getIconSmall(ICONS.K.DICTAPHONE_STOP));
		}
	}

	/**
	 * start/stop the dictaphone
	 *
	 * @param it
	 */
	public void dictaphoneStart(JButton it) {
		String cmd = getHomeDir().getAbsolutePath()
				+ File.separator + ".storybook5/dictaphone"
				+ File.separator + "nerd-dictation"
				+ File.separator + "nerd-dictation.py";
		if (dictaphoneStarted) {
			dictaphoneStarted = false;
			it.setToolTipText(I18N.getMsg("dictaphone.start"));
			it.setIcon(IconUtil.getIconSmall(ICONS.K.DICTAPHONE_START));
			CallProcess.runCommand("python3", cmd, "end");
		} else {
			dictaphoneStarted = true;
			it.setToolTipText(I18N.getMsg("dictaphone.stop"));
			it.setIcon(IconUtil.getIconSmall(ICONS.K.DICTAPHONE_STOP));
			CallProcess.runCommand("python3", cmd, "begin",
					"--continuous",
					"--full-sentence",
					"--punctuate-from-previous-timeout", "2",
					"--idle-time", "0.1"
			);
		}
	}

}
